import 'dart:io';
import 'package:billettapp_for_vitensenteret/resources/util.dart';
import 'package:billettapp_for_vitensenteret/admin_ticket.dart';
import 'package:billettapp_for_vitensenteret/settings.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:firebase_auth/firebase_auth.dart';
import 'package:image_picker/image_picker.dart';
import 'main.dart';
import 'tickets.dart';
import 'firebase_functions.dart';

bool showLoadingLogIn = false;
bool showLoadingRegister = false;


///AuthenticationPage - redirects the user to the correct page base on whether
///they are logged in or not
class AuthenticationPage extends StatelessWidget{
  const AuthenticationPage({super.key});

  @override
  Widget build(BuildContext context) {

    return Navigator(
      onGenerateRoute: (settings){
        Widget page = const LogInPage();
        if (FirebaseAuth.instance.currentUser != null) {
          //firebaseGetTickets();
          page = const TicketsPage();
          if (FirebaseAuth.instance.currentUser!.email!.contains(RegExp(r"^[a-zA-Z0-9.a-zA-Z0-9.!#$%&'*+-/=?^_`{|}~]+@vitensenteret.no")) && kIsWeb) {
            page = const AdminTicketPage();
          }
        }
        return MaterialPageRoute(builder: (_)=> page);
      },
    );
  }

}

///LogInPage - Displays the log in page, with a form for all log in information
///and a button for switching to the register page
class LogInPage extends StatefulWidget {
  const LogInPage( {super.key} );

  @override
  State<LogInPage> createState() => _LogInPageState();
}

class _LogInPageState extends State<LogInPage>
    with TickerProviderStateMixin{

  //Controller to handle animation
  late AnimationController _animationController;

  //Initialize controllers
  @override
  void initState() {
    _animationController = AnimationController(
      vsync: this,
      duration: const Duration(seconds: 5),
    )..addListener(() {
      setState(() {});
    });
    _animationController.repeat(reverse: false);

    super.initState();
  }

  //Dispose of controllers when class is destroyed
  @override
  void dispose() {
    _animationController.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {

    return Scaffold(
      body: Center(
        child: Column(
          mainAxisAlignment: MainAxisAlignment.center,
          children: [

            //Display the log in form
            Container(
              //height: 250,
              width: 350,
              alignment: Alignment.center,
              child: Card(
                child: Padding(
                  padding: const EdgeInsets.all(10.0),
                  child: Column (
                    children: const [
                      _LoginForm(),
                    ],
                  ),
                ),
              ),
            ),

            //Empty space
            const SizedBox(height: 5,),

            //Button that redirects to the register page
            TextButton(
                onPressed: () {
                  Navigator.push((context), MaterialPageRoute(builder: (context) => const RegisterPage()));
                },
                child: const Text("Registrer bruker")),

            //Loading circle, is shown when logging in is in progress
            Stack(
              alignment: Alignment.center,
              children: [
                //Container(child: ),
                if (showLoadingLogIn) CircularProgressIndicator(
                  value: _animationController.value,
                  semanticsLabel: 'Circular Progress Indicator',
                ),
              ],
            )

          ],
        ),
      ),
    );
  }
}


///loginForm - the text boxes and buttons associated with logging in
class _LoginForm extends StatefulWidget {
  const _LoginForm ();
  @override
  State<_LoginForm> createState() => _LoginFormState();
}

class _LoginFormState extends State<_LoginForm> {
  final GlobalKey<FormState> _formKey = GlobalKey<FormState>();

  var email = "";                   //email user is trying to log in with
  var password = "";                //password user is trying to log in with

  bool errorMessage = false;        //should the error message widget be displayed
  var feedback = 0;                 //return value from firebase function

  @override
  Widget build(BuildContext context) {
    return Form(
      key: _formKey,
      child: Column(
        children: <Widget> [

          //Email input
          TextFormField(
            decoration: const InputDecoration(
              hintText: 'Epost',
              icon: Icon(Icons.mail),
            ),
            validator: (String? value) {
              if (value == null || value.isEmpty) {
                return 'Skriv inn epost';
              }
              email = value;
              return null;
            },
          ),

          //Password input
          TextFormField(
            obscureText: true,
            decoration: const InputDecoration(
              hintText: 'Passord',
              icon: Icon(Icons.password),
            ),
            validator: (String? value) {
              if (value == null || value.isEmpty) {
                return 'Skriv inn passord';
              }
              password = value;
              return null;
            },
          ),

          //Error messages based on the feedback from firebase logging in function
          Stack(
            alignment: Alignment.center,
            children: [

              if (feedback == 1 && errorMessage) const Padding(
                padding: EdgeInsets.only(top: 15.0),
                child: Text(
                  'Feil brukernavn eller passord',
                  style: TextStyle(color: Colors.red),
                ),
              ),
              if (feedback == 2 && errorMessage) const Padding(
                padding: EdgeInsets.only(top: 15.0),
                child: Text (
                  'Feil under innlogging, prøv igjen senere',
                  style: TextStyle(color: Colors.red),
                ),
              )

            ],
          ),

          //Empty space
          const SizedBox(height: 15,),

            //Log in button, validates log in and redirects when accepted
            ElevatedButton(
              onPressed: () async {
                //If all the fields are validated correctly
                if (_formKey.currentState!.validate()) {
                  showLoadingLogIn = true;
                  feedback = await firebaseLogIn(password, email);

                  //check if log in succeeded by listening for changes to the authentication
                  FirebaseAuth.instance
                      .userChanges()
                      .listen((User? user) {
                    //If log in failed, hide loading symbol and show error message
                    if (user == null) {
                      showLoadingLogIn = false;
                      setState(() {
                        errorMessage = true;
                      });
                    } else {

                      if (FirebaseAuth.instance.currentUser!.email!.contains(RegExp(r"^[a-zA-Z0-9.a-zA-Z0-9.!#$%&'*+-/=?^_`{|}~]+@vitensenteret.no")) && kIsWeb && selectedIndex == 2) {
                        Navigator.push(context, MaterialPageRoute(builder: (context) => const AdminTicketPage()));
                      } else if (selectedIndex == 2) {
                        Navigator.push(context, MaterialPageRoute(builder: (context) => const TicketsPage()));
                      } else if (selectedIndex == 3) {
                        //if user logs in via profile page, they get redirected to profile page
                        Navigator.push(context, MaterialPageRoute(builder: (context) => ProfilePage()));
                      }
                    }
                  });

              }
            },
            child: const Text("Logg inn"),
          )
        ],
      ),
    );
  }
}


///RegisterPage - holds the register form, as well as a button to redirect
///to the log in page
class RegisterPage extends StatefulWidget {
  const RegisterPage( {super.key} );

  @override
  State<RegisterPage> createState() => _RegisterPageState();
}

class _RegisterPageState extends State<RegisterPage>
    with TickerProviderStateMixin{
  late AnimationController _animationController;

  //Initialize controllers
  @override
  void initState() {
    _animationController = AnimationController(
      vsync: this,
      duration: const Duration(seconds: 5),
    )..addListener(() {
      setState(() {});
    });
    _animationController.repeat(reverse: false);

    super.initState();
  }

  //Dispose of controllers when class is destroyed
  @override
  void dispose() {
    _animationController.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {

    return Scaffold(
      body: Center(
        child: SingleChildScrollView(
          child: Column(
            mainAxisAlignment: MainAxisAlignment.center,
            children: [

              //Register form
              Container(
                //height: 617,
                width: 350,
                alignment: Alignment.center,
                child: Card(
                  child: Padding(
                    padding: const EdgeInsets.all(10.0),
                    child: Column (

                      children: const [
                        _RegisterForm(),

                        SizedBox(height: 25,),

                      ],
                    ),
                  ),
                ),
              ),

              //Empty space
              const SizedBox(height: 15,),

              //Button to send user to log in page
              TextButton(
                  onPressed: () {
                    Navigator.push(context, MaterialPageRoute(builder: (context) => const LogInPage()));
                  },
                  child: const Text("Logg inn")
              ),

              //Loading animation
              Stack(
                alignment: Alignment.center,
                children: [
                  //Container(child: ),
                  if (showLoadingRegister) CircularProgressIndicator(
                    value: _animationController.value,
                    semanticsLabel: 'Circular Progress Indicator',
                  ),
                ],
              )

            ],
          ),
        ),
      ),
    );
  }
}

///RegisterForm - contains text boxes and buttons for registering a new user
class _RegisterForm extends StatefulWidget {
  const _RegisterForm();
  @override
  State<_RegisterForm> createState() => _RegisterFormState();
}

class _RegisterFormState extends State<_RegisterForm> {
  final GlobalKey<FormState> _formKey = GlobalKey<FormState>();
  final ImagePicker _picker = ImagePicker();

  DateTime currentDate = DateTime.now();
  var emailAddress = "";
  var password = "";
  var repeatPassword = "";
  var firstName = "";
  var lastName = "";
  var mobile = "";
  late File profilePic;                   //mobile and web profile pic uses two
  var webProfilePic;                      //different variables

  bool errorMessage = false;
  var feedback = 0;
  var imagePath = 'assets/default.png';
  bool imagePicked = false;

  /// Function for picking a date, gathered from:
  /// https://www.appsdeveloperblog.com/datepicker-example-in-flutter/
  Future<void> _selectDate(BuildContext context) async {
    final DateTime? pickedDate = await showDatePicker(
      context: context,
      initialDate: DateTime.now(),
      firstDate: DateTime(1920),
      lastDate: DateTime.now(),
    );
    if (pickedDate != null && pickedDate != currentDate) {
      setState(() {
        currentDate = pickedDate;
      });
    }
  }

  @override
  Widget build(BuildContext context) {
    return Form(
      key: _formKey,
      child: Column(
        children: <Widget> [
          //If profile pic is not chosen yet, show default pic
          if (!imagePicked)CircleAvatar(
            maxRadius: 70,
            backgroundImage: AssetImage(
              imagePath,
            ),
          ),

          //show picked profile pic
          if (imagePicked) CircleAvatar(
            maxRadius: 70,
            backgroundImage: FileImage(
              profilePic,
            ),
          ),

          //Button for opening image picker
          ElevatedButton(
              onPressed: () async {

                //For mobile images
                if (!kIsWeb) {
                  final XFile? image = await _picker.pickImage(source: ImageSource.gallery);

                  if (image != null) {
                    profilePic = File(image.path);
                    setState(() {
                      imagePicked = true;
                    });
                  }
                }
                //For web images
                else if (kIsWeb) {
                  XFile? image = await _picker.pickImage(source: ImageSource.gallery);

                  if (image != null){
                    var f = await image.readAsBytes();
                    webProfilePic = f;
                    setState(() {
                      imagePicked = true;
                    });
                  }
                }
              },
              child: const Text("Legg til bilde")),

          //Enter first name
          TextFormField(
            decoration: const InputDecoration(
              hintText: 'Fornavn',
              icon: Icon(Icons.person),
            ),
            validator: (String? value) {
              if (value == null || value.isEmpty) {
                return 'Skriv inn fornavn';
              }
              firstName = value;
              return null;
            },
          ),

          //Enter last name
          TextFormField(
            decoration: const InputDecoration(
              hintText: 'Etternavn',
              icon: Icon(Icons.person),
            ),
            validator: (String? value) {
              if (value == null || value.isEmpty) {
                return 'Skriv inn etternavn';
              }
              lastName = value;
              return null;
            },
          ),

          //Enter email
          TextFormField(
            keyboardType: TextInputType.emailAddress,
            decoration: const InputDecoration(
                hintText: 'Epost',
                icon: Icon(Icons.email)
            ),
            validator: (String? value) {
              if (value == null || value.isEmpty) {
                return 'Skriv inn epost';
              }
              if (!validateEmail(value)) {
                return 'Feil format på epost';
              }
              emailAddress = value;
              return null;
            },
          ),

          //Enter phone number
          TextFormField(
            keyboardType: TextInputType.number,
            decoration: const InputDecoration(
              hintText: 'Mobilnummer',
              icon: Icon(Icons.phone_android),
            ),
            validator: (String? value) {
              if (value == null || value.isEmpty) {
                return 'Skriv inn mobilnummer';
              }
              if (!validatePhone(value)) {
                return 'Ugyldig mobilnummer';
              }
              mobile = value;
              return null;
            },
          ),

          //Pick birth date
          Padding(
            padding: const EdgeInsets.only(top: 8.0),
            child: Row(
              children: [
                const Icon(Icons.date_range, color: Colors.grey,),
                const Padding(
                  padding: EdgeInsets.only(left: 15.0, right: 8.0),
                  child: Text(
                    "Fødselsdag:",
                    style: TextStyle(
                      fontSize: 17,
                      color: Colors.blueGrey,
                    ),),
                ),
                OutlinedButton(
                    onPressed: () {
                      _selectDate(context);
                    },
                    child: Text(
                      "${currentDate.day}/${currentDate.month}/${currentDate.year}",
                      style: const TextStyle(
                        fontSize: 17,
                      ),
                    )
                ),
              ],
            ),
          ),

          //Write password first time
          TextFormField(
            obscureText: true,
            decoration: const InputDecoration(
              hintText: 'Passord',
              icon: Icon(Icons.password),
            ),
            validator: (String? value) {
              if (value == null || value.isEmpty) {
                return 'Skriv in passord';
              }
              var ret = validatePassword(value);
              if (ret == 1) {
                return 'Passord må være minst 8 tegn langt';
              }
              if (ret == 2) {
                return 'Passord må inneholde minst en bokstav';
              }
              if (ret == 3) {
                return 'Passord må inneholde minst et tall';
              }
              password = value;
              return null;
            },
          ),

          //Write password a second time
          TextFormField(
            obscureText: true,
            decoration: const InputDecoration(
              hintText: 'Gjenta passord',
              icon: Icon(Icons.password),
            ),
            validator: (String? value) {
              if (value == null || value.isEmpty) {
                return 'Skriv in passord på nytt';
              }
              if (value != password) {
                return 'Passordet er ikke det samme';
              }
              repeatPassword = value;
              return null;
            },
          ),

          //Error messages that can be shown
          Stack(
            alignment: Alignment.center,
            children: [

              if (feedback == 2 && errorMessage) const Padding(
                padding: EdgeInsets.only(top: 15.0),
                child: Text(
                  'For svakt passord',
                  style: TextStyle(color: Colors.red),
                ),
              ),
              if (feedback == 3 && errorMessage) const Padding(
                padding: EdgeInsets.only(top: 15.0),
                child: Text (
                  'Bruker finnes allerede',
                  style: TextStyle(color: Colors.red),
                ),
              ),
              if (feedback == 1 && errorMessage) const Padding(
                padding: EdgeInsets.only(top: 15.0),
                child: Text (
                  'Feil under registrering, prøv igjen senere',
                  style: TextStyle(color: Colors.red),
                ),
              ),
              if (feedback == 4 && errorMessage) const Padding(
                padding: EdgeInsets.only(top: 15.0),
                child: Text (
                  'Bilde mangler',
                  style: TextStyle(color: Colors.red),
                ),
              ),

            ],
          ),

          //Empty space
          const SizedBox(height: 15,),

          //Register button
          ElevatedButton(
              onPressed: () async {
                //Checks if all fields in the form is validated
                if (_formKey.currentState!.validate() && imagePicked) {
                  showLoadingRegister = true;
                  feedback = await firebaseRegister(password, emailAddress);

                  //Check if user authentication changed
                  FirebaseAuth.instance
                      .idTokenChanges()
                      .listen((User? user) {
                    //If user is not authenticated
                    if (user == null) {
                      showLoadingRegister = false;
                      setState(() {
                        errorMessage = true;
                      });
                    } else {
                      //Send information about the user to the database
                      firebaseSendRegisterInfo("$firstName $lastName", emailAddress,
                          profilePic, FirebaseAuth.instance.currentUser!.uid, mobile,
                          "${currentDate.day}/${currentDate.month}/${currentDate.year}", profilePic);

                      if (selectedIndex == 2) {
                        //If user registered from ticket page, they get redirected to tickets page
                        Navigator.push(context, MaterialPageRoute(builder: (context) => const TicketsPage()));
                      } else if (selectedIndex == 3) {
                        //If user registered from profile page, they get redirected to profile page
                        Navigator.push(context, MaterialPageRoute(builder: (context) => ProfilePage()));
                      }
                    }
                  });

              } else {
                  setState(() {
                    errorMessage = true;
                    feedback = 4;
                  });
                }
            },
            child: const Text("Registrer"),
          )
        ],
      ),
    );
  }
}