import 'dart:math';

import 'package:billettapp_for_vitensenteret/colors.dart';
import 'package:flip_card/flip_card.dart';
import 'package:flutter/material.dart';
import 'package:billettapp_for_vitensenteret/home_games.dart';

import '../data/word_list.dart';
import '../models/letter_model.dart';
import '../models/word_model.dart';
import '../widgets/board.dart';
import '../widgets/keyboard.dart';

enum GameStatus {playing, submitting, lost, won}

/// Contains wordle game widget
class WordleScreen extends StatefulWidget {
  const WordleScreen({Key? key}) : super (key: key);

  @override
  _WordleScreenState createState() => _WordleScreenState();
}

class _WordleScreenState extends State<WordleScreen> {
  // Status for what is currently being done by the player
  GameStatus _gameStatus = GameStatus.playing;

  // Board filled with empty letters
  final List<Word> _board = List.generate(
      6,
      (_) => Word(letters: List.generate(5, (_) => Letter.empty())),
  );

  // Flip animation
  final List<List<GlobalKey<FlipCardState>>> _flipCardKeys = List.generate(
    6,
      (_) => List.generate(5,(_)=> GlobalKey<FlipCardState>()),
  );

  // Which index the user is on
  int _currentWordIndex = 0;

  // Current guessed word from the current row
  Word? get _currentWord =>
  _currentWordIndex < _board.length ? _board[_currentWordIndex] : null;


  // Solution
  Word _solution = Word.fromString(
    fiveLetterWords[Random().nextInt(fiveLetterWords.length)].toUpperCase(),
  );

  final Set<Letter> _keyboardLetters = {};

  final String gamedescription = "1. Målet er å gjette rett ord.\n"
      "2. Ordet har ikke reppeterende bokstaver.\n"
      "3. Ordet kan være norskt.\n"
      " Tips det kan ha noe med Vitensenteret å gjøre!\n\n"
      "                                            Lykke til!";

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      // Top bar
      appBar: AppBar(leading: BackButton(
        onPressed: () => Navigator.push(
          context,
          MaterialPageRoute(builder: (context) => const GameMenuPage()),
        ),
      ),
        title: const Text('Viten Ord'),
        titleSpacing: 00.0,
        centerTitle: true,
        toolbarHeight: 60.2,
        toolbarOpacity: 0.8,
        shape: const RoundedRectangleBorder(
          borderRadius: BorderRadius.only(
              bottomRight: Radius.circular(25),
              bottomLeft: Radius.circular(25)),
        ),
        elevation: 0.00,
        backgroundColor: morkBlaa,
      ),
      // Game window
      body: Column(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          Board(board: _board, flipCardKeys: _flipCardKeys),
          const SizedBox(height: 80),
          Keyboard(
            onKeyTapped: _onKeyTapped,
            onDeleteTapped: _onDeleteTapped,
            onEnterTapped: _onEnterTapped,
            letters: _keyboardLetters,
          ),
          // Bottom menu
          Row(
            mainAxisAlignment: MainAxisAlignment.center,
            children: [
              // Rules button
              TextButton(
              onPressed: () => showDialog<String>(
                context: context,
                builder: (BuildContext context) => AlertDialog(
                  title: const Text('Hvordan spille Viten Ord:'),
                  content: Text(gamedescription),
                  actions: const <Widget>[
                  ],
                ),
              ),
              style: TextButton.styleFrom(
                foregroundColor: morkBlaa,
              ),
              child: const Text('Regler'),
              ),
              // Restart
              TextButton(
                style: TextButton.styleFrom(
                  foregroundColor: morkBlaa,
                ),
                onPressed: _restart,
                child: const Text("Start på nytt"),
              ),
            ],
          ),
        ],
      ),
    );
  }

  // When a letter is tapped on
  void _onKeyTapped(String val) {
    if(_gameStatus == GameStatus.playing) {
      setState(() => _currentWord?.addLetter(val));
    }
  }

  // When the delete button is tapped
  void _onDeleteTapped() {
    if(_gameStatus == GameStatus.playing) {
      setState(() => _currentWord?.removeLetter());
    }
  }

  // When the enter button is tapped
  Future<void> _onEnterTapped() async{
    if(_gameStatus == GameStatus.playing &&
    _currentWord != null &&
    !_currentWord!.letters.contains(Letter.empty())) {
      _gameStatus = GameStatus.submitting;

      final letters = <Letter>[];
      for(var i = 0; i < _currentWord!.letters.length; i++){
        final currentWordLetter = _currentWord!.letters[i];
        final currentSolutionLetter = _solution.letters[i];

        // Update word with colors
        // Note that this means a guess with like aaaaa where a is in the word, will tell users that aaaaa are all in the word
        setState(() {
          if(currentWordLetter == currentSolutionLetter) {
            letters.add(currentWordLetter);
            _currentWord!.letters[i] =
                currentWordLetter.copyWith(status: LetterStatus.correct);

          } else if(_solution.letters.contains(currentWordLetter)) {
            letters.add(currentWordLetter);


              // Check if guess has placed this correctly, before flipping it to yellow
              bool guessedRightAnotherTime = false;
              int latestIndex = 0;
              for (int j = 0; j < _currentWord!.letters.length; j++) {
                if (_currentWord!.letters[j].val == _solution.letters[j].val &&
                    _currentWord!.letters[j].val == currentWordLetter.val) {
                  guessedRightAnotherTime = true;
                }
                if (_currentWord!.letters[j].val == currentWordLetter.val) {
                  latestIndex = j;
                }
              }

              // If its guessed correctly later, no need to mark it
              // However if it is guessed later, flip the latest "closest" to the right guess
              if (!guessedRightAnotherTime && latestIndex == i) {
                _currentWord!.letters[i] =
                    currentWordLetter.copyWith(status: LetterStatus.inWord);
              }else {
                _currentWord!.letters[i] =
                    currentWordLetter.copyWith(status: LetterStatus.notInWord);
              }
          }

          else {
            _currentWord!.letters[i] =
                currentWordLetter.copyWith(status: LetterStatus.notInWord);
          }
        });

        final letter = _keyboardLetters.firstWhere(
            (e)=> e.val == currentWordLetter.val,
          orElse: () => Letter.empty(),
        );
        if (letter.status != LetterStatus.correct) {
          _keyboardLetters.removeWhere((e)=>e.val == currentWordLetter.val);
          _keyboardLetters.add(_currentWord!.letters[i]);
        }

        await Future.delayed(
          const Duration(milliseconds: 150),
            () => _flipCardKeys[_currentWordIndex][i].currentState?.toggleCard(),
        );
      }
      _checkIfWinOrLoss();
    }
  }



  /// Checks if current guessed word is the solution word
  void _checkIfWinOrLoss(){
    if(_currentWord!.wordString == _solution.wordString) {
      _gameStatus = GameStatus.won; //if win
      ScaffoldMessenger.of(context).showSnackBar(
        SnackBar(
          dismissDirection: DismissDirection.none,
          duration: const Duration(days: 1),
          backgroundColor: gronn,
          content: const Text(
            'Du vant!',
            style: TextStyle(color: Colors.white),
          ),
          action: SnackBarAction(
            onPressed: _restart, //let user start new game
            textColor: Colors.white,
            label: 'Nytt spill',
          )
        ),
      );
    } else if (_currentWordIndex + 1 >= _board.length){
        _gameStatus = GameStatus.lost; //Change gamestate
        ScaffoldMessenger.of(context).showSnackBar(
          SnackBar(
              dismissDirection: DismissDirection.none,
              duration: const Duration(days: 1), //make this last for a day
              backgroundColor: gronn,
              content: Text(
                "Du tapte! ${_solution.wordString}",
                style: const TextStyle(color: Colors.white),
              ),
              action: SnackBarAction(
                onPressed: _restart, //let user start new game
                textColor: Colors.white,
                label: 'Nytt spill',
              )
          ),
        );
    } else {
      _gameStatus = GameStatus.playing;
    }
    _currentWordIndex += 1;
  }

  /// Resets values and fetches a new solution word
  void _restart() {
    setState(() {
      _gameStatus = GameStatus.playing;
      _currentWordIndex = 0;
      _board
      ..clear()
      ..addAll(
      List.generate(
          6,
            (_) => Word(letters: List.generate(5, (_) => Letter.empty())),
        ),
      );
      _solution = Word.fromString(
        fiveLetterWords[Random().nextInt(fiveLetterWords.length)].toUpperCase(),
      );
      _flipCardKeys
        ..clear()
        ..addAll(
        List.generate(
          6,
            (_) => List.generate(5, (_)=> GlobalKey<FlipCardState>()),
        ),
      );
      _keyboardLetters.clear();
    });
  }
}

/// Returns the index where the instance of the [item] was located at in [suggestion]
int firstIndexFoundAt(List<Letter> suggestion, Letter item){
  for (var i = 0; i < suggestion.length; i++){
    if (suggestion[i] == item){
      return i;
    }
  }
  return 10;
}

/// Checks if there are more than one instances of an [item] in the [suggestion]
bool guessContainsMultipleLetters(List<Letter> suggestion, Letter item) {
  var count = 1;
  for (var i = 0; i < suggestion.length; i++){
    if (suggestion[i] == item){
      count ++;
    }
  }
  if (count > 1){
    return true;
  }
  return false;
}